@@ -8,9 +8,9 @@
  ************************************************************************/
 #ifdef RCS
 static /*const*/char rcsid[]=
- "$Id: formail.c,v 1.62 1994/08/18 13:44:52 berg Exp $";
+ "$Id: formail.c,v 1.73 1994/10/31 17:30:23 berg Exp $";
 #endif
-static /*const*/char rcsdate[]="$Date: 1994/08/18 13:44:52 $";
+static /*const*/char rcsdate[]="$Date: 1994/10/31 17:30:23 $";
 #include "includes.h"
 #include <ctype.h>		/* iscntrl() */
 #include "formail.h"
@@ -35,7 +35,7 @@ static const char
  x_[]=			"X-",				/* general extension */
  old_[]=		OLD_PREFIX,			     /* my extension */
  xloop[]=		"X-Loop:",				/* ditto ... */
- unknown[]=UNKNOWN,re[]=" Re:",fmusage[]=FM_USAGE;
+ daemon[]="<>",unknown[]=UNKNOWN,re[]=" Re:",fmusage[]=FM_USAGE;
 
 static const struct {const char*hedr;int lnr;}cdigest[]=
 {
@@ -89,7 +89,7 @@ static const char emboxsep[]=eMAILBOX_SEPARATOR;
 
 const char binsh[]=BinSh,sfolder[]=FOLDER,
  couldntw[]="Couldn't write to stdout";
-int errout,oldstdout,quiet=1,buflast,lenfileno;
+int errout,oldstdout,quiet=1,zap,buflast,lenfileno;
 long initfileno;
 char ffileno[LEN_FILENO_VAR+8*sizeof(initfileno)*4/10+1+1]=DEFfileno;
 int lexitcode;					     /* dummy, for waitfor() */
@@ -132,13 +132,14 @@ static void renfield(pointer,oldl,newname,newl)struct field**const pointer;
   if(newname[newl-1]==HEAD_DELIMITER)		     /* completely new field */
 replaceall:
      oldl=p->id_len;			     /* replace the old one entirely */
-  p->fld_text[p->tot_len-1]='\n';p->tot_len=(i=p->tot_len-oldl)+newl;
+  p->id_len+=(int)newl-(int)oldl;p->fld_text[p->tot_len-1]='\n';
+  p->tot_len=(i=p->tot_len-oldl)+newl;
   if(newl>oldl)
      *pointer=p=realloc(p,FLD_HEADSIZ+p->tot_len);
   chp=p->fld_text;tmemmove(chp+newl,chp+oldl,i);tmemmove(chp,newname,newl);
 }
 
-static void procfields P((void))
+static void procfields(sareply)const int sareply;
 { struct field*fldp,**afldp;
   fldp= *(afldp= &rdheader);
   while(fldp)
@@ -148,13 +149,15 @@ static void procfields P((void))
 	 fp2->tot_len-fp2->id_len);
 	goto fixfldp;
       }
-     if((fp2=findf(fldp,&iheader))&&
-	!(areply&&fldp->id_len>=fp2->tot_len-1))
+     if(!sareply&&
+	(fp2=findf(fldp,&iheader))&&
+	!(areply&&fldp->id_len>=fp2->tot_len-1))      /* filled replacement? */
       { renfield(afldp,(size_t)0,old_,STRLEN(old_));	/* implicitly rename */
 fixfldp:
 	fldp= *afldp;
       }
-     if(findf(fldp,&Iheader))				    /* delete fields */
+     if((fp2=findf(fldp,&Iheader))&&			    /* delete fields */
+	!(sareply&&fldp->id_len<fp2->tot_len-1))       /* empty replacement? */
 	goto delfld;
      ;{ struct field*uf;
 	if((uf=findf(fldp,&uheader))&&!uf->fld_ref)
@@ -192,6 +195,97 @@ static int artheadr P((void))	     /* could it be the start of an article? */
    }
   return 0;
 }
+			     /* lifted out of main() to reduce main()'s size */
+static char*getsender(namep,fldp,trust)char*namep;struct field*fldp;
+ const int trust;
+{ char*chp;int i,nowm;size_t j;static int lastm;
+  chp=fldp->fld_text;j=fldp->id_len;i=maxindex(sest);
+  while((sest[i].len!=j||strnIcmp(sest[i].head,chp,j))&&i--);
+  if(i>=0&&(i!=maxindex(sest)||fldp==rdheader))		  /* found anything? */
+   { char*saddr;char*tmp;			     /* determine the weight */
+     nowm=trust?sest[i].wtrepl:areply?sest[i].wrepl:i;chp+=j;
+     tmp=malloc(j=fldp->tot_len-j);tmemmove(tmp,chp,j);(chp=tmp)[j-1]='\0';
+     if(sest[i].head==From_)
+      { char*pastad;
+	if(trust||!(saddr=strchr(chp,'\n')))	     /* skip the first line? */
+	   saddr=chp;						  /* no need */
+	if(*saddr=='\n'&&(pastad=strchr(saddr,' ')))
+	   saddr=pastad+1;			/* reposition at the address */
+	chp=saddr;
+	while((pastad=strchr(chp,'\n'))&&(pastad=strchr(pastad,' ')))
+	   chp=pastad+1;		      /* skip to the last uucp >From */
+	if(pastad=strchr(chp,' '))			/* found an address? */
+	 { char*savetmp;				      /* lift it out */
+	   savetmp=malloc((j=pastad-chp)+1);tmemmove(savetmp,chp,j);
+	   savetmp[j]='\0';
+	   if(strchr(savetmp,'@'))			 /* domain attached? */
+	      chp=savetmp,savetmp=tmp,tmp=chp;			/* ok, ready */
+	   else					/* no domain, bang away! :-) */
+	    { static const char remf[]=" remote from ",fwdb[]=" forwarded by ";
+	      char*p1,*p2;
+	      chp=tmp;
+	      for(;;)
+	       { int c;
+		 p1=strstr(saddr,remf);
+		 if(!(p2=strstr(saddr,fwdb))&&!p1)
+		    break;				     /* no more info */
+		 if(!p1||p2&&p2<p1)		      /* pick the first bang */
+		    p1=p2+STRLEN(fwdb);
+		 else
+		    p1+=STRLEN(remf);
+		 for(;;)				     /* copy it over */
+		  { switch(c= *p1++)
+		     { default:*chp++=c;
+			  continue;
+		       case '\0':case '\n':*chp++='!';	     /* for the buck */
+		     }
+		    break;
+		  }
+		 saddr=p1;				/* continue the hunt */
+	       }
+	      strcpy(chp,savetmp);chp=tmp;	     /* attach the user part */
+	    }
+	   free(savetmp);	  /* (temporary buffers might have switched) */
+	 }
+      }
+     while(*(chp=skpspace(chp))=='\n')
+	chp++;
+     for(saddr=0;;chp=skipwords(chp))			/* skip RFC 822 wise */
+      { switch(*chp)
+	 { default:
+	      if(!saddr)		   /* if we haven't got anything yet */
+		 saddr=chp;			/* this might be the address */
+	      continue;
+	   case '<':skipwords(saddr=chp);	  /* hurray, machine useable */
+	   case '\0':;
+	 }
+	break;
+      }
+     if(saddr)				    /* any useful mailaddress found? */
+      { if(*saddr)				  /* did it have any length? */
+	 { if(!strpbrk(saddr,"@!/"))
+	      nowm-=(maxindex(sest)+2)*4;		/* depreciate "user" */
+	   else if(strstr(saddr,".UUCP"))
+	      nowm-=(maxindex(sest)+2)*3;	 /* depreciate .UUCP address */
+	   else if(strchr(saddr,'@')&&!strchr(saddr,'.'))
+	      nowm-=(maxindex(sest)+2)*2;	     /* depreciate user@host */
+	   else if(strchr(saddr,'!'))
+	      nowm-=(maxindex(sest)+2)*1;	     /* depreciate bangpaths */
+	   if(!namep||nowm>lastm)		/* better than previous ones */
+	      goto pnewname;
+	 }
+	else if(sest[i].head==returnpath)		/* nill Return-Path: */
+	 { saddr=(char*)daemon;nowm=maxindex(sest)+2;		 /* override */
+pnewname:  lastm=nowm;saddr=strcpy(malloc(strlen(saddr)+1),saddr);
+	   if(namep)
+	      free(namep);
+	   namep=saddr;
+	 }
+      }
+     free(tmp);
+   }					   /* save headers for later perusal */
+  return namep;
+}
 
 static PROGID;
 
@@ -228,6 +322,8 @@ main(lastm,argv)int lastm;const char*const argv[];
 		 continue;
 	      case FM_CONCATENATE:conctenate=1;
 		 continue;
+	      case FM_ZAPWHITE:zap=1;
+		 continue;
 	      case FM_QUIET:quiet=1;
 		 if(*chp=='-')
 		    chp++,quiet=0;
@@ -416,122 +512,55 @@ startover:
      do rex[i].rexl=0;
      while(i--);			      /* reset all state information */
      clear_uhead(uheader);clear_uhead(Uheader);
-     wasafrom_=!force&&rdheader&&eqFrom_(rdheader->fld_text);procfields();
-     for(fldp=rdheader;fldp;fldp=fldp->fld_next)    /* go through the linked */
-      { int nowm;				    /* list of header-fields */
+     wasafrom_=!force&&rdheader&&eqFrom_(rdheader->fld_text);
+     procfields(areply);
+     for(fldp= *(afldp= &rdheader);fldp;)
+      { if(zap)		      /* go through the linked list of header-fields */
+	 { chp=fldp->fld_text+(j=fldp->id_len);
+	   if(chp[-1]==HEAD_DELIMITER)
+	      if(*chp!=' '&&fldp->tot_len>j+1)
+	       { chp=j+(*afldp=fldp=
+		  realloc(fldp,FLD_HEADSIZ+(i= ++fldp->tot_len)))->fld_text;
+		 tmemmove(chp+1,chp,i-j);*chp=' ';
+	       }
+	      else if(fldp->tot_len<=j+2)
+	       { *afldp=fldp->fld_next;free(fldp);fldp= *afldp;
+		 continue;
+	       }
+	 }
 	if(conctenate)
-	   concatenate(fldp);			 /* look for `sender' fields */
-	chp=fldp->fld_text;j=fldp->id_len;i=maxindex(sest);
-	while((sest[i].len!=j||strnIcmp(sest[i].head,chp,j))&&i--);
-	if(i>=0&&(i!=maxindex(sest)||fldp==rdheader))	  /* found anything? */
-	 { char*saddr;char*tmp;			     /* determine the weight */
-	   nowm=trust?sest[i].wtrepl:areply?sest[i].wrepl:i;chp+=j;
-	   tmp=malloc(j=fldp->tot_len-j);tmemmove(tmp,chp,j);
-	   (chp=tmp)[j-1]='\0';
-	   if(sest[i].head==From_)
-	    { char*pastad;
-	      if(trust||!(saddr=strchr(chp,'\n')))   /* skip the first line? */
-		 saddr=chp;					  /* no need */
-	      if(*saddr=='\n'&&(pastad=strchr(saddr,' ')))
-		 saddr=pastad+1;		/* reposition at the address */
-	      chp=saddr;
-	      while((pastad=strchr(chp,'\n'))&&(pastad=strchr(pastad,' ')))
-		 chp=pastad+1;		      /* skip to the last uucp >From */
-	      if(pastad=strchr(chp,' '))		/* found an address? */
-	       { char*savetmp;				      /* lift it out */
-		 savetmp=malloc((j=pastad-chp)+1);tmemmove(savetmp,chp,j);
-		 savetmp[j]='\0';
-		 if(strchr(savetmp,'@'))		 /* domain attached? */
-		    chp=savetmp,savetmp=tmp,tmp=chp;		/* ok, ready */
-		 else				/* no domain, bang away! :-) */
-		  { static const char remf[]=" remote from ",
-		     fwdb[]=" forwarded by ";
-		    char*p1,*p2;
-		    chp=tmp;
-		    for(;;)
-		     { int c;
-		       p1=strstr(saddr,remf);
-		       if(!(p2=strstr(saddr,fwdb))&&!p1)
-			  break;			     /* no more info */
-		       if(!p1||p2&&p2<p1)	      /* pick the first bang */
-			  p1=p2+STRLEN(fwdb);
-		       else
-			  p1+=STRLEN(remf);
-		       for(;;)				     /* copy it over */
-			{ switch(c= *p1++)
-			   { default:*chp++=c;
-				continue;
-			     case '\0':case '\n':*chp++='!'; /* for the buck */
-			   }
-			  break;
-			}
-		       saddr=p1;			/* continue the hunt */
-		     }
-		    strcpy(chp,savetmp);chp=tmp;     /* attach the user part */
-		  }
-		 free(savetmp);	  /* (temporary buffers might have switched) */
-	       }
-	    }
-	   while(*(chp=skpspace(chp))=='\n')
-	      chp++;
-	   for(saddr=0;;chp=skipwords(chp))		/* skip RFC 822 wise */
-	    { switch(*chp)
-	       { default:
-		    if(!saddr)		   /* if we haven't got anything yet */
-		       saddr=chp;		/* this might be the address */
-		    continue;
-		 case '<':skipwords(saddr=chp);	  /* hurray, machine useable */
-		 case '\0':;
-	       }
-	      break;
-	    }
-	   if(saddr)			    /* any useful mailaddress found? */
-	    { if(*saddr)			  /* did it have any length? */
-	       { if(!strpbrk(saddr,"@!/"))
-		    nowm-=(maxindex(sest)+2)*4;		/* depreciate "user" */
-		 else if(strstr(saddr,".UUCP"))
-		    nowm-=(maxindex(sest)+2)*3;	 /* depreciate .UUCP address */
-		 else if(strchr(saddr,'@')&&!strchr(saddr,'.'))
-		    nowm-=(maxindex(sest)+2)*2;	     /* depreciate user@host */
-		 else if(strchr(saddr,'!'))
-		    nowm-=(maxindex(sest)+2)*1;	     /* depreciate bangpaths */
-		 if(!namep||nowm>lastm)		/* better than previous ones */
-		  { saddr=strcpy(malloc(strlen(saddr)+1),saddr);lastm=nowm;
-		    goto pnewname;
-		  }
-	       }
-	      else if(sest[i].head==returnpath)		/* nill Return-Path: */
-	       { saddr=0;lastm=maxindex(sest)+2;		 /* override */
-pnewname:	 if(namep)
-		    free(namep);
-		 namep=saddr;
-	       }
-	    }
-	   free(tmp);
-	 }				   /* save headers for later perusal */
-	i=maxindex(rex);chp=fldp->fld_text;j=fldp->id_len;    /* e.g. areply */
+	   concatenate(fldp);		    /* save fields for later perusal */
+	namep=getsender(namep,fldp,trust);i=maxindex(rex);chp=fldp->fld_text;
+	j=fldp->id_len;
 	while((rex[i].lenr!=j||strnIcmp(rex[i].headr,chp,j))&&i--);
 	chp+=j;
 	if(i>=0&&(j=fldp->tot_len-j)>1)			  /* found anything? */
 	 { tmemmove(rex[i].rexp=realloc(rex[i].rexp,(rex[i].rexl=j)+1),chp,j);
 	   rex[i].rexp[j]='\0';			     /* add a terminating \0 */
 	 }
+	fldp= *(afldp= &fldp->fld_next);
       }
      if(idcache)
-      { int dupid=0;
-	if(msid->rexl)					/* any Message-ID: ? */
-	 { insoffs=maxlen;msid->rexp[msid->rexl-1]='\0';
+      { int dupid=0;char*key;
+	key=namep;
+	if(!areply)
+	 { key=0;
+	   if(msid->rexl)				/* any Message-ID: ? */
+	      (key=msid->rexp)[msid->rexl-1]='\0';
+	 }
+	if(key)
+	 { insoffs=maxlen;
 	   do					/* wipe out trailing newline */
 	    { int j;char*p;	  /* start reading & comparing the next word */
-	      for(p=msid->rexp;(j=fgetc(idcache))==*p;p++)
+	      for(p=key;(j=fgetc(idcache))==*p;p++)
 		 if(!j)					     /* end of word? */
 		  { if(!quiet)
-		       nlog("Duplicate ID found:"),elog(msid->rexp),elog("\n");
+		       nlog("Duplicate key found:"),elog(key),elog("\n");
 		    dupid=1;
 		    goto dupfound;		     /* YES! duplicate found */
 		  }
 	      if(!j)					     /* end of word? */
-	       { if(p==msid->rexp&&insoffs==maxlen)	 /* first character? */
+	       { if(p==key&&insoffs==maxlen)		 /* first character? */
 		  { insoffs=ftell(idcache)-1;		     /* found end of */
 		    goto skiprest;			  /* circular buffer */
 		  }
@@ -551,10 +580,11 @@ skiprest:	 for(;;)			/* skip the rest of the word */
 	   while(ftell(idcache)<maxlen);		  /* past our quota? */
 noluck:	   if(insoffs>=maxlen)				  /* past our quota? */
 	      insoffs=0;			     /* start up front again */
-	   fseek(idcache,insoffs,SEEK_SET);
-	   fwrite(msid->rexp,1,msid->rexl+1,idcache);
+	   fseek(idcache,insoffs,SEEK_SET);fwrite(key,1,strlen(key)+1,idcache);
+	   putc('\0',idcache);			   /* mark new end of buffer */
 dupfound:  fseek(idcache,(off_t)0,SEEK_SET);	 /* rewind, for any next run */
-	   msid->rexp[msid->rexl-1]='\n';	      /* restore the newline */
+	   if(!areply)
+	      key[msid->rexl-1]='\n';		      /* restore the newline */
 	 }
 	if(!split)			  /* not splitting?  terminate early */
 	   return dupid?EXIT_SUCCESS:1;
@@ -600,7 +630,7 @@ dupfound:  fseek(idcache,(off_t)0,SEEK_SET);	 /* rewind, for any next run */
 	 }
 	if(msid->rexl)			 /* do we add an In-Reply-To: field? */
 	   loadbuf(inreplyto,STRLEN(inreplyto)),loadsaved(msid),addbuf();
-	procfields();
+	procfields(0);
       }
      else if(!force&&		       /* are we allowed to add From_ lines? */
 	     (!rdheader||!eqFrom_(rdheader->fld_text))&&   /* is it missing? */
@@ -675,7 +705,7 @@ dupfound:  fseek(idcache,(off_t)0,SEEK_SET);	 /* rewind, for any next run */
   if(babyl)
    { int c,lc;					/* ditch pseudo BABYL header */
      for(lc=0;c=getchar(),c!=EOF&&(c!='\n'||lc!='\n');lc=c);
-     babylstart=0;
+     buflast=c;babylstart=0;
    }
   if(ctlength>0)
    { if(buffilled)
